@@ -1,4 +1,4 @@
-/* Copyright (c) 2016, 2021, MariaDB Corporation.
+/* Copyright (c) 2016, 2022, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -38,7 +38,7 @@
 /*
   Compare ASCII string against the string with the specified
   character set.
-  Only compares the equality, case insencitive.
+  Only compares the equality, case insensitive.
 */
 static bool eq_ascii_string(const CHARSET_INFO *cs,
                             const char *ascii,
@@ -133,7 +133,7 @@ static const char tab_arr[TAB_SIZE_LIMIT+1]= "        ";
 
 static int append_tab(String *js, int depth, int tab_size)
 {
-  if (js->append("\n", 1))
+  if (js->append('\n'))
     return 1;
   for (int i=0; i<depth; i++)
   {
@@ -257,16 +257,21 @@ static int json_nice(json_engine_t *je, String *nice_js,
                      Item_func_json_format::formats mode, int tab_size=4)
 {
   int depth= 0;
-  const char *comma, *colon;
+  static const char *comma= ", ", *colon= "\": ";
   uint comma_len, colon_len;
   int first_value= 1;
+  int value_size = 0;
+  int curr_state= -1;
+  int64_t value_len= 0;
+  String curr_str{};
+
+  nice_js->length(0);
+  nice_js->set_charset(je->s.cs);
+  nice_js->alloc(je->s.str_end - je->s.c_str + 32);
 
-  DBUG_ASSERT(je->s.cs == nice_js->charset());
   DBUG_ASSERT(mode != Item_func_json_format::DETAILED ||
               (tab_size >= 0 && tab_size <= TAB_SIZE_LIMIT));
 
-  comma= ", ";
-  colon= "\": ";
   if (mode == Item_func_json_format::LOOSE)
   {
     comma_len= 2;
@@ -285,6 +290,7 @@ static int json_nice(json_engine_t *je, String *nice_js,
 
   do
   {
+    curr_state= je->state;
     switch (je->state)
     {
     case JST_KEY:
@@ -307,7 +313,7 @@ static int json_nice(json_engine_t *je, String *nice_js,
             append_tab(nice_js, depth, tab_size))
           goto error;
 
-        nice_js->append("\"", 1);
+        nice_js->append('"');
         append_simple(nice_js, key_start, key_end - key_start);
         nice_js->append(colon, colon_len);
       }
@@ -332,17 +338,23 @@ handle_value:
         if (append_simple(nice_js, je->value_begin,
                           je->value_end - je->value_begin))
           goto error;
-
+        
+        curr_str.copy((const char *)je->value_begin,
+                      je->value_end - je->value_begin, je->s.cs);
+        value_len= je->value_end - je->value_begin;
         first_value= 0;
+        if (value_size != -1)
+          value_size++;
       }
       else
       {
         if (mode == Item_func_json_format::DETAILED &&
-            depth > 0 &&
+            depth > 0 && !(curr_state != JST_KEY) &&
             append_tab(nice_js, depth, tab_size))
           goto error;
         nice_js->append((je->value_type == JSON_VALUE_OBJECT) ? "{" : "[", 1);
         first_value= 1;
+        value_size= (je->value_type == JSON_VALUE_OBJECT) ? -1: 0;
         depth++;
       }
 
@@ -351,11 +363,27 @@ handle_value:
     case JST_OBJ_END:
     case JST_ARRAY_END:
       depth--;
-      if (mode == Item_func_json_format::DETAILED &&
+      if (mode == Item_func_json_format::DETAILED && (value_size > 1 || value_size == -1) &&
           append_tab(nice_js, depth, tab_size))
         goto error;
+        
+      if (mode == Item_func_json_format::DETAILED && 
+          value_size == 1 && je->state != JST_OBJ_END)
+      {
+        for (auto i = 0; i < value_len; i++)
+        {
+          nice_js->chop();
+        }
+        for (auto i = 0; i < (depth + 1) * tab_size + 1; i++)
+        {
+          nice_js->chop();
+        }
+        nice_js->append(curr_str);
+      }
+      
       nice_js->append((je->state == JST_OBJ_END) ? "}": "]", 1);
       first_value= 0;
+      value_size= -1;
       break;
 
     default:
@@ -363,7 +391,7 @@ handle_value:
     };
   } while (json_scan_next(je) == 0);
 
-  return je->s.error;
+  return je->s.error || *je->killed_ptr;
 
 error:
   return 1;
@@ -371,15 +399,15 @@ error:
 
 
 #define report_json_error(js, je, n_param) \
-  report_json_error_ex(js, je, func_name(), n_param, \
+  report_json_error_ex(js->ptr(), je, func_name(), n_param, \
       Sql_condition::WARN_LEVEL_WARN)
 
-void report_json_error_ex(String *js, json_engine_t *je,
+void report_json_error_ex(const char *js, json_engine_t *je,
                           const char *fname, int n_param,
                           Sql_condition::enum_warning_level lv)
 {
   THD *thd= current_thd;
-  int position= (int)((const char *) je->s.c_str - js->ptr());
+  int position= (int)((const char *) je->s.c_str - js);
   uint code;
 
   n_param++;
@@ -409,16 +437,22 @@ void report_json_error_ex(String *js, json_engine_t *je,
 
   case JE_DEPTH:
     code= ER_JSON_DEPTH;
-    push_warning_printf(thd, lv, code, ER_THD(thd, code), JSON_DEPTH_LIMIT,
-                        n_param, fname, position);
+    if (lv == Sql_condition::WARN_LEVEL_ERROR)
+      my_error(code, MYF(0), JSON_DEPTH_LIMIT, n_param, fname, position);
+    else
+      push_warning_printf(thd, lv, code, ER_THD(thd, code), JSON_DEPTH_LIMIT,
+                          n_param, fname, position);
     return;
 
   default:
     return;
   }
 
-  push_warning_printf(thd, lv, code, ER_THD(thd, code),
-                      n_param, fname, position);
+  if (lv == Sql_condition::WARN_LEVEL_ERROR)
+    my_error(code, MYF(0), n_param, fname, position);
+  else
+    push_warning_printf(thd, lv, code, ER_THD(thd, code),
+                        n_param, fname, position);
 }
 
 
@@ -428,15 +462,15 @@ void report_json_error_ex(String *js, json_engine_t *je,
 #define TRIVIAL_PATH_NOT_ALLOWED 3
 
 #define report_path_error(js, je, n_param) \
-  report_path_error_ex(js, je, func_name(), n_param,\
+  report_path_error_ex(js->ptr(), je, func_name(), n_param,\
       Sql_condition::WARN_LEVEL_WARN)
 
-static void report_path_error_ex(String *ps, json_path_t *p,
-                                 const char *fname, int n_param,
-                                 Sql_condition::enum_warning_level lv)
+void report_path_error_ex(const char *ps, json_path_t *p,
+                          const char *fname, int n_param,
+                          Sql_condition::enum_warning_level lv)
 {
   THD *thd= current_thd;
-  int position= (int)((const char *) p->s.c_str - ps->ptr() + 1);
+  int position= (int)((const char *) p->s.c_str - ps + 1);
   uint code;
 
   n_param++;
@@ -455,8 +489,11 @@ static void report_path_error_ex(String *ps, json_path_t *p,
 
   case JE_DEPTH:
     code= ER_JSON_PATH_DEPTH;
-    push_warning_printf(thd, lv, code, ER_THD(thd, code),
-                        JSON_DEPTH_LIMIT, n_param, fname, position);
+    if (lv == Sql_condition::WARN_LEVEL_ERROR)
+      my_error(code, MYF(0), JSON_DEPTH_LIMIT, n_param, fname, position);
+    else
+      push_warning_printf(thd, lv, code, ER_THD(thd, code),
+                          JSON_DEPTH_LIMIT, n_param, fname, position);
     return;
 
   case NO_WILDCARD_ALLOWED:
@@ -471,12 +508,14 @@ static void report_path_error_ex(String *ps, json_path_t *p,
   default:
     return;
   }
-  push_warning_printf(thd, lv, code, ER_THD(thd, code),
-                      n_param, fname, position);
+  if (lv == Sql_condition::WARN_LEVEL_ERROR)
+    my_error(code, MYF(0), n_param, fname, position);
+  else
+    push_warning_printf(thd, lv, code, ER_THD(thd, code),
+                        n_param, fname, position);
 }
 
 
-
 /*
   Checks if the path has '.*' '[*]' or '**' constructions
   and sets the NO_WILDCARD_ALLOWED error if the case.
@@ -510,7 +549,7 @@ bool Item_func_json_exists::fix_length_and_dec()
 {
   if (Item_bool_func::fix_length_and_dec())
     return TRUE;
-  maybe_null= 1;
+  set_maybe_null();
   path.set_constant_flag(args[1]->const_item());
   return FALSE;
 }
@@ -564,7 +603,7 @@ bool Item_func_json_value::fix_length_and_dec()
   collation.set(args[0]->collation);
   max_length= args[0]->max_length;
   set_constant_flag(args[1]->const_item());
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -574,15 +613,11 @@ bool Item_func_json_query::fix_length_and_dec()
   collation.set(args[0]->collation);
   max_length= args[0]->max_length;
   set_constant_flag(args[1]->const_item());
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
 
-/*
-  Returns NULL, not an error if the found value
-  is not a scalar.
-*/
 bool Json_path_extractor::extract(String *str, Item *item_js, Item *item_jp,
                                   CHARSET_INFO *cs)
 {
@@ -615,6 +650,9 @@ continue_search:
   if (json_read_value(&je))
     return true;
 
+  if (je.value_type == JSON_VALUE_NULL)
+    return true;
+
   if (unlikely(check_and_get_value(&je, str, &error)))
   {
     if (error)
@@ -704,9 +742,9 @@ String *Item_func_json_quote::val_str(String *str)
   str->length(0);
   str->set_charset(&my_charset_utf8mb4_bin);
 
-  if (str->append("\"", 1) ||
+  if (str->append('"') ||
       st_append_escaped(str, s) ||
-      str->append("\"", 1))
+      str->append('"'))
   {
     /* Report an error. */
     null_value= 1;
@@ -722,7 +760,7 @@ bool Item_func_json_unquote::fix_length_and_dec()
   collation.set(&my_charset_utf8mb3_general_ci,
                 DERIVATION_COERCIBLE, MY_REPERTOIRE_ASCII);
   max_length= args[0]->max_length;
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -843,7 +881,7 @@ bool Item_func_json_extract::fix_length_and_dec()
   max_length= args[0]->max_length * (arg_count - 1);
 
   mark_constant_paths(paths, args+1, arg_count-1);
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -918,7 +956,7 @@ String *Item_func_json_extract::read_json(String *str,
     str->set_charset(js->charset());
     str->length(0);
 
-    if (possible_multiple_values && str->append("[", 1))
+    if (possible_multiple_values && str->append('['))
       goto error;
   }
 
@@ -966,7 +1004,7 @@ String *Item_func_json_extract::read_json(String *str,
     if (!possible_multiple_values)
     {
       /* Loop to the end of the JSON just to make sure it's valid. */
-      while (json_get_path_next(&je, &p) == 0) {}
+      while (json_scan_next(&je) == 0) {}
       break;
     }
   }
@@ -980,14 +1018,12 @@ String *Item_func_json_extract::read_json(String *str,
     goto return_null;
   }
 
-  if (possible_multiple_values && str->append("]", 1))
+  if (possible_multiple_values && str->append(']'))
     goto error; /* Out of memory. */
 
   js= str;
   json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
-  tmp_js.length(0);
-  tmp_js.set_charset(js->charset());
   if (json_nice(&je, &tmp_js, Item_func_json_format::LOOSE))
     goto error;
 
@@ -1099,7 +1135,6 @@ my_decimal *Item_func_json_extract::val_decimal(my_decimal *to)
       case JSON_VALUE_ARRAY:
       case JSON_VALUE_FALSE:
       case JSON_VALUE_UNINITALIZED:
-      // TODO: fix: NULL should be NULL
       case JSON_VALUE_NULL:
         int2my_decimal(E_DEC_FATAL_ERROR, 0, false/*unsigned_flag*/, to);
         return to;
@@ -1115,7 +1150,7 @@ bool Item_func_json_contains::fix_length_and_dec()
 {
   a2_constant= args[1]->const_item();
   a2_parsed= FALSE;
-  maybe_null= 1;
+  set_maybe_null();
   if (arg_count > 2)
     path.set_constant_flag(args[2]->const_item());
   return Item_bool_func::fix_length_and_dec();
@@ -1248,7 +1283,7 @@ static int check_contains(json_engine_t *js, json_engine_t *value)
     if (value->value_type != JSON_VALUE_STRING)
       return FALSE;
     /*
-       TODO: make proper json-json comparison here that takes excapint
+       TODO: make proper json-json comparison here that takes excipient
              into account.
      */
     return value->value_len == js->value_len &&
@@ -1374,7 +1409,7 @@ bool Item_func_json_contains_path::fix_length_and_dec()
 {
   ooa_constant= args[1]->const_item();
   ooa_parsed= FALSE;
-  maybe_null= 1;
+  set_maybe_null();
   mark_constant_paths(paths, args+2, arg_count-2);
   return Item_bool_func::fix_length_and_dec();
 }
@@ -1536,7 +1571,7 @@ longlong Item_func_json_contains_path::val_int()
     n_found= arg_count - 2;
   }
   else
-    n_found= 0; /* Jost to prevent 'uninitialized value' warnings */
+    n_found= 0; /* Just to prevent 'uninitialized value' warnings */
 
   result= 0;
   while (json_get_path_next(&je, &p) == 0)
@@ -1586,14 +1621,14 @@ null_return:
   `CONVERT(arg USING charset)` is actually a general purpose string
   expression, not a JSON expression.
 */
-static bool is_json_type(const Item *item)
+bool is_json_type(const Item *item)
 {
   for ( ; ; )
   {
     if (Type_handler_json_common::is_json_type_handler(item->type_handler()))
       return true;
     const Item_func_conv_charset *func;
-    if (!(func= dynamic_cast<const Item_func_conv_charset*>(item)))
+    if (!(func= dynamic_cast<const Item_func_conv_charset*>(item->real_item())))
       return false;
     item= func->arguments()[0];
   }
@@ -1634,15 +1669,15 @@ static int append_json_value(String *str, Item *item, String *tmp_val)
 
     if (item->result_type() == STRING_RESULT)
     {
-      return str->append("\"", 1) ||
+      return str->append('"') ||
              st_append_escaped(str, sv) ||
-             str->append("\"", 1);
+             str->append('"');
     }
     return st_append_escaped(str, sv);
   }
 
 append_null:
-  return str->append("null", 4);
+  return str->append(STRING_WITH_LEN("null"));
 }
 
 
@@ -1680,15 +1715,15 @@ static int append_json_value_from_field(String *str,
 
     if (i->result_type() == STRING_RESULT)
     {
-      return str->append("\"", 1) ||
+      return str->append('"') ||
              st_append_escaped(str, sv) ||
-             str->append("\"", 1);
+             str->append('"');
     }
     return st_append_escaped(str, sv);
   }
 
 append_null:
-  return str->append("null", 4);
+  return str->append(STRING_WITH_LEN("null"));
 }
 
 
@@ -1698,7 +1733,7 @@ static int append_json_keyname(String *str, Item *item, String *tmp_val)
   if (item->null_value)
     goto append_null;
 
-  return str->append("\"", 1) ||
+  return str->append('"') ||
          st_append_escaped(str, sv) ||
          str->append("\": ", 3);
 
@@ -1728,7 +1763,7 @@ bool Item_func_json_array::fix_length_and_dec()
     return TRUE;
 
   for (n_arg=0 ; n_arg < arg_count ; n_arg++)
-    char_length+= args[n_arg]->max_char_length() + 4;
+    char_length+= static_cast<ulonglong>(args[n_arg]->max_char_length()) + 4;
 
   fix_char_length_ulonglong(char_length);
   tmp_val.set_charset(collation.collation);
@@ -1738,13 +1773,13 @@ bool Item_func_json_array::fix_length_and_dec()
 
 String *Item_func_json_array::val_str(String *str)
 {
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
   uint n_arg;
 
   str->length(0);
   str->set_charset(collation.collation);
 
-  if (str->append("[", 1) ||
+  if (str->append('[') ||
       ((arg_count > 0) && append_json_value(str, args[0], &tmp_val)))
     goto err_return;
 
@@ -1755,7 +1790,7 @@ String *Item_func_json_array::val_str(String *str)
       goto err_return;
   }
 
-  if (str->append("]", 1))
+  if (str->append(']'))
     goto err_return;
 
   if (result_limit == 0)
@@ -1787,11 +1822,12 @@ bool Item_func_json_array_append::fix_length_and_dec()
   for (n_arg= 1; n_arg < arg_count; n_arg+= 2)
   {
     paths[n_arg/2].set_constant_flag(args[n_arg]->const_item());
-    char_length+= args[n_arg/2+1]->max_char_length() + 4;
+    char_length+=
+        static_cast<ulonglong>(args[n_arg+1]->max_char_length()) + 4;
   }
 
   fix_char_length_ulonglong(char_length);
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -1803,8 +1839,9 @@ String *Item_func_json_array_append::val_str(String *str)
   uint n_arg, n_path;
   size_t str_rest_len;
   const uchar *ar_end;
+  THD *thd= current_thd;
 
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
 
   if ((null_value= args[0]->null_value))
     return 0;
@@ -1830,6 +1867,7 @@ String *Item_func_json_array_append::val_str(String *str)
 
     json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                     (const uchar *) js->ptr() + js->length());
+    je.killed_ptr= (uchar*)&thd->killed;
 
     c_path->cur_step= c_path->p.steps;
 
@@ -1884,11 +1922,11 @@ String *Item_func_json_array_append::val_str(String *str)
       else
         c_to= je.value_end;
 
-      if (str->append("[", 1) ||
+      if (str->append('[') ||
           str->append((const char *) c_from, c_to - c_from) ||
           str->append(", ", 2) ||
           append_json_value(str, args[n_arg+1], &tmp_val) ||
-          str->append("]", 1) ||
+          str->append(']') ||
           str->append((const char *) je.s.c_str,
                       js->end() - (const char *) je.s.c_str))
         goto return_null; /* Out of memory. */
@@ -1910,8 +1948,7 @@ String *Item_func_json_array_append::val_str(String *str)
 
   json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
-  str->length(0);
-  str->set_charset(js->charset());
+  je.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je, str, Item_func_json_format::LOOSE))
     goto js_error;
 
@@ -1921,6 +1958,7 @@ js_error:
   report_json_error(js, &je, 0);
 
 return_null:
+  thd->check_killed(); // to get the error message right
   null_value= 1;
   return 0;
 }
@@ -1931,8 +1969,9 @@ String *Item_func_json_array_insert::val_str(String *str)
   json_engine_t je;
   String *js= args[0]->val_json(&tmp_js);
   uint n_arg, n_path;
+  THD *thd= current_thd;
 
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
 
   if ((null_value= args[0]->null_value))
     return 0;
@@ -1968,6 +2007,7 @@ String *Item_func_json_array_insert::val_str(String *str)
 
     json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                     (const uchar *) js->ptr() + js->length());
+    je.killed_ptr= (uchar*)&thd->killed;
 
     c_path->cur_step= c_path->p.steps;
 
@@ -2007,7 +2047,7 @@ String *Item_func_json_array_insert::val_str(String *str)
         goto js_error;
     }
 
-    if (unlikely(je.s.error))
+    if (unlikely(je.s.error || *je.killed_ptr))
       goto js_error;
 
     str->length(0);
@@ -2051,8 +2091,7 @@ String *Item_func_json_array_insert::val_str(String *str)
 
   json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
-  str->length(0);
-  str->set_charset(js->charset());
+  je.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je, str, Item_func_json_format::LOOSE))
     goto js_error;
 
@@ -2061,6 +2100,7 @@ String *Item_func_json_array_insert::val_str(String *str)
 js_error:
   report_json_error(js, &je, 0);
 return_null:
+  thd->check_killed(); // to get the error message right
   null_value= 1;
   return 0;
 }
@@ -2068,13 +2108,13 @@ return_null:
 
 String *Item_func_json_object::val_str(String *str)
 {
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
   uint n_arg;
 
   str->length(0);
   str->set_charset(collation.collation);
 
-  if (str->append("{", 1) ||
+  if (str->append('{') ||
       (arg_count > 0 &&
        (append_json_keyname(str, args[0], &tmp_val) ||
         append_json_value(str, args[1], &tmp_val))))
@@ -2088,7 +2128,7 @@ String *Item_func_json_object::val_str(String *str)
       goto err_return;
   }
 
-  if (str->append("}", 1))
+  if (str->append('}'))
     goto err_return;
 
   if (result_limit == 0)
@@ -2135,7 +2175,7 @@ static int do_merge(String *str, json_engine_t *je1, json_engine_t *je2)
   
     json_string_set_cs(&key_name, je1->s.cs);
 
-    if (str->append("{", 1))
+    if (str->append('{'))
       return 3;
     while (json_scan_next(je1) == 0 &&
            je1->state != JST_OBJ_END)
@@ -2161,7 +2201,7 @@ static int do_merge(String *str, json_engine_t *je1, json_engine_t *je2)
         *je2= sav_je2;
       }
 
-      if (str->append("\"", 1) ||
+      if (str->append('"') ||
           append_simple(str, key_start, key_end - key_start) ||
           str->append("\":", 2))
         return 3;
@@ -2229,8 +2269,7 @@ merged_j1:
             return 2;
           continue;
         }
-        if (json_skip_key(je2) ||
-            json_skip_level(je1))
+        if (json_skip_key(je2) || json_skip_level(je1))
           return 1;
         goto continue_j2;
       }
@@ -2246,7 +2285,7 @@ merged_j1:
       if (json_skip_key(je2))
         return 1;
 
-      if (str->append("\"", 1) ||
+      if (str->append('"') ||
           append_simple(str, key_start, je2->s.c_str - key_start))
         return 3;
 
@@ -2254,7 +2293,7 @@ continue_j2:
       continue;
     }
 
-    if (str->append("}", 1))
+    if (str->append('}'))
       return 3;
   }
   else
@@ -2274,7 +2313,7 @@ continue_j2:
     }
     else
     {
-      if (str->append("[", 1))
+      if (str->append('['))
         return 3;
       if (je1->value_type == JSON_VALUE_OBJECT)
       {
@@ -2316,7 +2355,7 @@ continue_j2:
       return 3;
 
     if (je2->value_type != JSON_VALUE_ARRAY &&
-        str->append("]", 1))
+        str->append(']'))
       return 3;
   }
 
@@ -2326,10 +2365,11 @@ continue_j2:
 
 String *Item_func_json_merge::val_str(String *str)
 {
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
   json_engine_t je1, je2;
   String *js1= args[0]->val_json(&tmp_js1), *js2=NULL;
   uint n_arg;
+  THD *thd= current_thd;
   LINT_INIT(js2);
 
   if (args[0]->null_value)
@@ -2346,9 +2386,11 @@ String *Item_func_json_merge::val_str(String *str)
 
     json_scan_start(&je1, js1->charset(),(const uchar *) js1->ptr(),
                     (const uchar *) js1->ptr() + js1->length());
+    je1.killed_ptr= (uchar*)&thd->killed;
 
     json_scan_start(&je2, js2->charset(),(const uchar *) js2->ptr(),
                     (const uchar *) js2->ptr() + js2->length());
+    je2.killed_ptr= (uchar*)&thd->killed;
 
     if (do_merge(str, &je1, &je2))
       goto error_return;
@@ -2370,8 +2412,7 @@ String *Item_func_json_merge::val_str(String *str)
 
   json_scan_start(&je1, js1->charset(),(const uchar *) js1->ptr(),
                   (const uchar *) js1->ptr() + js1->length());
-  str->length(0);
-  str->set_charset(js1->charset());
+  je1.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je1, str, Item_func_json_format::LOOSE))
     goto error_return;
 
@@ -2383,6 +2424,7 @@ error_return:
     report_json_error(js1, &je1, 0);
   if (je2.s.error)
     report_json_error(js2, &je2, n_arg);
+  thd->check_killed(); // to get the error message right
 null_return:
   null_value= 1;
   return NULL;
@@ -2415,7 +2457,7 @@ static int copy_value_patch(String *str, json_engine_t *je)
   }
   /* JSON_VALUE_OBJECT */
 
-  if (str->append("{", 1))
+  if (str->append('{'))
     return 1;
   while (json_scan_next(je) == 0 && je->state != JST_OBJ_END)
   {
@@ -2438,12 +2480,12 @@ static int copy_value_patch(String *str, json_engine_t *je)
     else
       first_key= 0;
 
-    if (str->append("\"", 1) ||
+    if (str->append('"') ||
         append_simple(str, key_start, je->value_begin - key_start) ||
         copy_value_patch(str, je))
       return 1;
   }
-  if (str->append("}", 1))
+  if (str->append('}'))
     return 1;
 
   return 0;
@@ -2479,7 +2521,7 @@ static int do_merge_patch(String *str, json_engine_t *je1, json_engine_t *je2,
     *empty_result= FALSE;
     json_string_set_cs(&key_name, je1->s.cs);
 
-    if (str->append("{", 1))
+    if (str->append('{'))
       return 3;
     while (json_scan_next(je1) == 0 &&
            je1->state != JST_OBJ_END)
@@ -2505,7 +2547,7 @@ static int do_merge_patch(String *str, json_engine_t *je1, json_engine_t *je2,
         *je2= sav_je2;
       }
 
-      if (str->append("\"", 1) ||
+      if (str->append('"') ||
           append_simple(str, key_start, key_end - key_start) ||
           str->append("\":", 2))
         return 3;
@@ -2596,7 +2638,7 @@ merged_j1:
       if (!first_key && str->append(", ", 2))
         return 3;
 
-      if (str->append("\"", 1) ||
+      if (str->append('"') ||
           append_simple(str, key_start, key_end - key_start) ||
           str->append("\":", 2))
         return 3;
@@ -2617,7 +2659,7 @@ continue_j2:
       continue;
     }
 
-    if (str->append("}", 1))
+    if (str->append('}'))
       return 3;
   }
   else
@@ -2636,11 +2678,12 @@ continue_j2:
 
 String *Item_func_json_merge_patch::val_str(String *str)
 {
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
   json_engine_t je1, je2;
   String *js1= args[0]->val_json(&tmp_js1), *js2=NULL;
   uint n_arg;
   bool empty_result, merge_to_null;
+  THD *thd= current_thd;
 
   /* To report errors properly if some JSON is invalid. */
   je1.s.error= je2.s.error= 0;
@@ -2657,6 +2700,7 @@ String *Item_func_json_merge_patch::val_str(String *str)
 
     json_scan_start(&je2, js2->charset(),(const uchar *) js2->ptr(),
                     (const uchar *) js2->ptr() + js2->length());
+    je2.killed_ptr= (uchar*)&thd->killed;
 
     if (merge_to_null)
     {
@@ -2676,12 +2720,13 @@ String *Item_func_json_merge_patch::val_str(String *str)
 
     json_scan_start(&je1, js1->charset(),(const uchar *) js1->ptr(),
                     (const uchar *) js1->ptr() + js1->length());
+    je1.killed_ptr= (uchar*)&thd->killed;
 
     if (do_merge_patch(str, &je1, &je2, &empty_result))
       goto error_return;
 
     if (empty_result)
-      str->append("null");
+      str->append(STRING_WITH_LEN("null"));
 
 cont_point:
     {
@@ -2704,8 +2749,7 @@ cont_point:
 
   json_scan_start(&je1, js1->charset(),(const uchar *) js1->ptr(),
                   (const uchar *) js1->ptr() + js1->length());
-  str->length(0);
-  str->set_charset(js1->charset());
+  je1.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je1, str, Item_func_json_format::LOOSE))
     goto error_return;
 
@@ -2717,6 +2761,7 @@ error_return:
     report_json_error(js1, &je1, 0);
   if (je2.s.error)
     report_json_error(js2, &je2, n_arg);
+  thd->check_killed(); // to get the error message right
 null_return:
   null_value= 1;
   return NULL;
@@ -2727,7 +2772,7 @@ bool Item_func_json_length::fix_length_and_dec()
 {
   if (arg_count > 1)
     path.set_constant_flag(args[1]->const_item());
-  maybe_null= 1;
+  set_maybe_null();
   max_length= 10;
   return FALSE;
 }
@@ -2873,7 +2918,7 @@ bool Item_func_json_type::fix_length_and_dec()
 {
   collation.set(&my_charset_utf8mb3_general_ci);
   max_length= 12;
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -2938,11 +2983,12 @@ bool Item_func_json_insert::fix_length_and_dec()
   for (n_arg= 1; n_arg < arg_count; n_arg+= 2)
   {
     paths[n_arg/2].set_constant_flag(args[n_arg]->const_item());
-    char_length+= args[n_arg/2+1]->max_char_length() + 4;
+    char_length+=
+        static_cast<ulonglong>(args[n_arg+1]->max_char_length()) + 4;
   }
 
   fix_char_length_ulonglong(char_length);
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -2953,8 +2999,9 @@ String *Item_func_json_insert::val_str(String *str)
   String *js= args[0]->val_json(&tmp_js);
   uint n_arg, n_path;
   json_string_t key_name;
+  THD *thd= current_thd;
 
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
 
   if ((null_value= args[0]->null_value))
     return 0;
@@ -2993,6 +3040,7 @@ String *Item_func_json_insert::val_str(String *str)
 
     json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                     (const uchar *) js->ptr() + js->length());
+    je.killed_ptr= (uchar*)&thd->killed;
 
     if (c_path->p.last_step < c_path->p.steps)
       goto v_found;
@@ -3042,7 +3090,7 @@ String *Item_func_json_insert::val_str(String *str)
         str->length(0);
         /* Wrap the value as an array. */
         if (append_simple(str, js->ptr(), (const char *) v_from - js->ptr()) ||
-            (do_array_autowrap && str->append("[", 1)))
+            (do_array_autowrap && str->append('[')))
           goto js_error; /* Out of memory. */
 
         if (je.value_type == JSON_VALUE_OBJECT)
@@ -3055,7 +3103,7 @@ String *Item_func_json_insert::val_str(String *str)
              (append_simple(str, v_from, je.s.c_str - v_from) ||
               str->append(", ", 2))) ||
             append_json_value(str, args[n_arg+1], &tmp_val) ||
-            (do_array_autowrap && str->append("]", 1)) ||
+            (do_array_autowrap && str->append(']')) ||
             append_simple(str, je.s.c_str, js->end()-(const char *) je.s.c_str))
           goto js_error; /* Out of memory. */
 
@@ -3126,7 +3174,7 @@ String *Item_func_json_insert::val_str(String *str)
       str->length(0);
       if (append_simple(str, js->ptr(), v_to - js->ptr()) ||
           (n_key > 0 && str->append(", ", 2)) ||
-          str->append("\"", 1) ||
+          str->append('"') ||
           append_simple(str, lp->key, lp->key_end - lp->key) ||
           str->append("\":", 2) ||
           append_json_value(str, args[n_arg+1], &tmp_val) ||
@@ -3174,7 +3222,7 @@ continue_point:
 
   json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
-  str->length(0);
+  je.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je, str, Item_func_json_format::LOOSE))
     goto js_error;
 
@@ -3182,6 +3230,7 @@ continue_point:
 
 js_error:
   report_json_error(js, &je, 0);
+  thd->check_killed(); // to get the error message right
 return_null:
   null_value= 1;
   return 0;
@@ -3194,7 +3243,7 @@ bool Item_func_json_remove::fix_length_and_dec()
   max_length= args[0]->max_length;
 
   mark_constant_paths(paths, args+1, arg_count-1);
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -3205,8 +3254,9 @@ String *Item_func_json_remove::val_str(String *str)
   String *js= args[0]->val_json(&tmp_js);
   uint n_arg, n_path;
   json_string_t key_name;
+  THD *thd= current_thd;
 
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
 
   if (args[0]->null_value)
     goto null_return;
@@ -3251,6 +3301,7 @@ String *Item_func_json_remove::val_str(String *str)
 
     json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                     (const uchar *) js->ptr() + js->length());
+    je.killed_ptr= (uchar*)&thd->killed;
 
     c_path->cur_step= c_path->p.steps;
 
@@ -3359,8 +3410,7 @@ v_found:
 
   json_scan_start(&je, js->charset(),(const uchar *) js->ptr(),
                   (const uchar *) js->ptr() + js->length());
-  str->length(0);
-  str->set_charset(js->charset());
+  je.killed_ptr= (uchar*)&thd->killed;
   if (json_nice(&je, str, Item_func_json_format::LOOSE))
     goto js_error;
 
@@ -3368,6 +3418,7 @@ v_found:
   return str;
 
 js_error:
+  thd->check_killed(); // to get the error message right
   report_json_error(js, &je, 0);
 null_return:
   null_value= 1;
@@ -3379,7 +3430,7 @@ bool Item_func_json_keys::fix_length_and_dec()
 {
   collation.set(args[0]->collation);
   max_length= args[0]->max_length;
-  maybe_null= 1;
+  set_maybe_null();
   if (arg_count > 1)
     path.set_constant_flag(args[1]->const_item());
   return FALSE;
@@ -3471,7 +3522,7 @@ skip_search:
     goto null_return;
   
   str->length(0);
-  if (str->append("[", 1))
+  if (str->append('['))
     goto err_return; /* Out of memory. */
   /* Parse the OBJECT collecting the keys. */
   while (json_scan_next(&je) == 0 && je.state != JST_OBJ_END)
@@ -3494,9 +3545,9 @@ skip_search:
       if (!check_key_in_list(str, key_start, key_len))
       { 
         if ((n_keys > 0 && str->append(", ", 2)) ||
-          str->append("\"", 1) ||
+          str->append('"') ||
           append_simple(str, key_start, key_len) ||
-          str->append("\"", 1))
+          str->append('"'))
         goto err_return;
         n_keys++;
       }
@@ -3511,7 +3562,7 @@ skip_search:
     }
   }
 
-  if (unlikely(je.s.error || str->append("]", 1)))
+  if (unlikely(je.s.error || str->append(']')))
     goto err_return;
 
   null_value= 0;
@@ -3549,7 +3600,7 @@ bool Item_func_json_search::fix_length_and_dec()
 
   /*
     It's rather difficult to estimate the length of the result.
-    I belive arglen^2 is the reasonable upper limit.
+    I believe arglen^2 is the reasonable upper limit.
   */
   if (args[0]->max_length > SQR_MAX_BLOB_WIDTH)
     max_length= MAX_BLOB_WIDTH;
@@ -3564,7 +3615,7 @@ bool Item_func_json_search::fix_length_and_dec()
 
   if (arg_count > 4)
     mark_constant_paths(paths, args+4, arg_count-4);
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -3615,14 +3666,14 @@ static int append_json_path(String *str, const json_path_t *p)
     else /*JSON_PATH_ARRAY*/
     {
 
-      if (str->append("[", 1) ||
+      if (str->append('[') ||
           str->append_ulonglong(c->n_item) ||
-          str->append("]", 1))
+          str->append(']'))
         return TRUE;
     }
   }
 
-  return str->append("\"", 1);
+  return str->append('"');
 }
 
 
@@ -3683,7 +3734,7 @@ String *Item_func_json_search::val_str(String *str)
         {
           if (n_path_found == 2)
           {
-            if (str->append("[", 1) ||
+            if (str->append('[') ||
                 append_json_path(str, &sav_path))
                 goto js_error;
           }
@@ -3709,7 +3760,7 @@ end:
   }
   else
   {
-    if (str->append("]", 1))
+    if (str->append(']'))
       goto js_error;
   }
 
@@ -3725,21 +3776,21 @@ null_return:
 }
 
 
-const char *Item_func_json_format::func_name() const
+LEX_CSTRING Item_func_json_format::func_name_cstring() const
 {
   switch (fmt)
   {
   case COMPACT:
-    return "json_compact";
+    return { STRING_WITH_LEN("json_compact") };
   case LOOSE:
-    return "json_loose";
+    return { STRING_WITH_LEN("json_loose") };
   case DETAILED:
-    return "json_detailed";
+    return { STRING_WITH_LEN("json_detailed") };
   default:
     DBUG_ASSERT(0);
   };
 
-  return "";
+  return NULL_clex_str;
 }
 
 
@@ -3748,7 +3799,7 @@ bool Item_func_json_format::fix_length_and_dec()
   decimals= 0;
   collation.set(args[0]->collation);
   max_length= args[0]->max_length;
-  maybe_null= 1;
+  set_maybe_null();
   return FALSE;
 }
 
@@ -3758,6 +3809,7 @@ String *Item_func_json_format::val_str(String *str)
   String *js= args[0]->val_json(&tmp_js);
   json_engine_t je;
   int tab_size= 4;
+  THD *thd= current_thd;
 
   if ((null_value= args[0]->null_value))
     return 0;
@@ -3781,13 +3833,13 @@ String *Item_func_json_format::val_str(String *str)
 
   json_scan_start(&je, js->charset(), (const uchar *) js->ptr(),
                   (const uchar *) js->ptr()+js->length());
+  je.killed_ptr= (uchar*)&thd->killed;
 
-  str->length(0);
-  str->set_charset(js->charset());
   if (json_nice(&je, str, fmt, tab_size))
   {
     null_value= 1;
     report_json_error(js, &je, 0);
+    thd->check_killed(); // to get the error message right
     return 0;
   }
 
@@ -3947,7 +3999,7 @@ Item_func_json_objectagg(THD *thd, Item_func_json_objectagg *item)
 {
   quick_group= FALSE;
   result.set_charset(collation.collation);
-  result.append("{");
+  result.append('{');
 }
 
 
@@ -3955,14 +4007,14 @@ bool
 Item_func_json_objectagg::fix_fields(THD *thd, Item **ref)
 {
   uint i;                       /* for loop variable */
-  DBUG_ASSERT(fixed == 0);
+  DBUG_ASSERT(fixed() == 0);
 
   memcpy(orig_args, args, sizeof(Item*) * arg_count);
 
   if (init_sum_func_check(thd))
     return TRUE;
 
-  maybe_null= 1;
+  set_maybe_null();
 
   /*
     Fix fields for select list and ORDER clause
@@ -3972,9 +4024,7 @@ Item_func_json_objectagg::fix_fields(THD *thd, Item **ref)
   {
     if (args[i]->fix_fields_if_needed_for_scalar(thd, &args[i]))
       return TRUE;
-    m_with_subquery|= args[i]->with_subquery();
-    with_param|= args[i]->with_param;
-    with_window_func|= args[i]->with_window_func;
+    with_flags|= args[i]->with_flags;
   }
 
   /* skip charset aggregation for order columns */
@@ -3991,7 +4041,7 @@ Item_func_json_objectagg::fix_fields(THD *thd, Item **ref)
   if (check_sum_func(thd, ref))
     return TRUE;
 
-  fixed= 1;
+  base_flags|= item_base_t::FIXED;
   return FALSE;
 }
 
@@ -4030,11 +4080,11 @@ bool Item_func_json_objectagg::add()
 
   null_value= 0;
   if (result.length() > 1)
-    result.append(", ");
+    result.append(STRING_WITH_LEN(", "));
 
-  result.append("\"");
-  result.append(*key);
-  result.append("\":");
+  result.append('"');
+  st_append_escaped(&result,key);
+  result.append(STRING_WITH_LEN("\":"));
 
   buf.length(0);
   append_json_value(&result, args[1], &buf);
@@ -4045,11 +4095,11 @@ bool Item_func_json_objectagg::add()
 
 String* Item_func_json_objectagg::val_str(String* str)
 {
-  DBUG_ASSERT(fixed == 1);
+  DBUG_ASSERT(fixed());
   if (null_value)
     return 0;
 
-  result.append("}");
+  result.append('}');
   return &result;
 }
 
